@@ -2,7 +2,7 @@
 #
 #  The MIT License
 #  
-#  Copyright (c) 2017-2018, 2020 Genome Research Ltd.
+#  Copyright (c) 2017 Genome Research Ltd.
 #  
 #  Author: Petr Danecek <pd3@sanger.ac.uk>
 #  
@@ -51,15 +51,12 @@ sub error
         "Options:\n",
         "   -a, --af-annots <file>      Allele frequency annotations [1000GP-AFs/AFs.tab.gz]\n",
         "   -i, --indir <dir>           Input directory with VCF files\n",
-        "       --include <expr>        Select sites for which the expression is true\n",
-        "       --exclude <expr>        Exclude sites for which the epxression is true\n",
         "   -l, --min-length <num>      Filter input regions shorter than this [1e6]\n",
         "   -m, --genmap <dir>          Directory with genetic map in IMPUTE2 format (optional)\n",
         "   -M, --rec-rate <float>      constant recombination rate per bp (optional)\n",
         "   -n, --min-markers <num>     Filter input regions with fewer marker than this [100]\n",
         "   -o, --outdir <dir>          Output directory\n",
         "   -q, --min-qual <num>        Filter input regions with quality smaller than this [10]\n",
-        "       --roh-args <string>     Extra arguments to pass to bcftools roh\n",
         "   -s, --silent                Quiet output, do not print commands\n",
         "   -h, -?, --help              This help message\n",
         "\n";
@@ -74,13 +71,9 @@ sub parse_params
         min_length  => 1e6,
         min_markers => 100,
         min_qual    => 10,
-        roh_args    => '',
     };
     while (defined(my $arg=shift(@ARGV)))
     {
-        if (                 $arg eq '--roh-args' ) { $$opts{roh_args}=shift(@ARGV); next }
-        if (                 $arg eq '--include' ) { $$opts{include_expr}=shift(@ARGV); next }
-        if (                 $arg eq '--exclude' ) { $$opts{exclude_expr}=shift(@ARGV); next }
         if ( $arg eq '-q' || $arg eq '--min-qual' ) { $$opts{min_qual}=shift(@ARGV); next }
         if ( $arg eq '-l' || $arg eq '--min-length' ) { $$opts{min_length}=shift(@ARGV); next }
         if ( $arg eq '-n' || $arg eq '--min-markers' ) { $$opts{min_markers}=shift(@ARGV); next }
@@ -99,8 +92,6 @@ sub parse_params
     if ( ! -e "$$opts{af_annots}.tbi" ) { error("The annotation file is not indexed: $$opts{af_annots}.tbi\n"); }
     if ( ! -e "$$opts{af_annots}.hdr" ) { error("The annotation file has no header: $$opts{af_annots}.hdr\n"); }
     if ( exists($$opts{genmap}) && ! -d "$$opts{genmap}" ) { error("The directory with genetic maps does not exist: $$opts{genmap}\n"); }
-    if ( exists($$opts{include_expr}) ) { $$opts{include_expr} =~ s/\'/\'\\\'\'/g; $$opts{inc_exc} .= qq[ -i '$$opts{include_expr}']; }
-    if ( exists($$opts{exclude_expr}) ) { $$opts{exclude_expr} =~ s/\'/\'\\\'\'/g; $$opts{inc_exc} .= qq[ -e '$$opts{exclude_expr}']; }
     return $opts;
 }
 
@@ -127,7 +118,7 @@ sub cmd
     else 
     {      
         # child
-        exec('bash', '-o','pipefail','-c', $cmd) or error("Failed to run the command [bash -o pipefail -c $cmd]: $!");
+        exec('/bin/bash', '-o','pipefail','-c', $cmd) or error("Failed to run the command [/bin/sh -o pipefail -c $cmd]: $!");
     }
 
     if ( exists($args{exit_on_error}) && !$args{exit_on_error} ) { return @out; }
@@ -180,16 +171,9 @@ sub parse_genmap_path
     }
     my @test = glob("$prefix*$suffix");
     if ( @test != @files ) 
-    {
-        error(
-            "Error: Could not determine the genetic map files in \"$$opts{genmap}\". The directory must contain only\n" .
-            "       the genetic map files (and nothing else) and the expected file names are\n" .
-            "           <prefix><chromosome-name><suffix>\n" .
-            "       for example:\n" .
-            "           genetic_map_chr10_combined_b37.txt\n" .
-            "           genetic_map_chr11_combined_b37.txt\n" .
-            "           ...\n" .
-            "       The heuristically determined wildcard \"$prefix*$suffix\" matches ",scalar @test," files, in total there are ",scalar @files," files.\n");
+    { 
+        print STDERR "Warning: Could not determine the genetic map files [$prefix][$suffix]\n";
+        return '';
     }
     return "-m $prefix\{CHROM}$suffix";
 }
@@ -212,19 +196,10 @@ sub run_roh
         my $outfile = "$$opts{outdir}/$`.bcf";
         push @files,$outfile;
         if ( -e $outfile ) { next; }
-
-        my $cmd = 
+        cmd(
             "bcftools annotate --rename-chrs $chr_fname '$$opts{indir}/$file' -Ou | " .
-            "bcftools annotate -c CHROM,POS,REF,ALT,AF1KG -h $$opts{af_annots}.hdr -a $$opts{af_annots} ";
-
-        if ( exists($$opts{inc_exc}) )
-        {
-            $cmd .= " -Ou | bcftools view $$opts{inc_exc} ";
-        }
-        $cmd .= "-Ob -o $outfile.part && ";
-        $cmd .= "mv $outfile.part $outfile";
-
-        cmd($cmd, %$opts);
+            "bcftools annotate -c CHROM,POS,REF,ALT,AF1KG -h $$opts{af_annots}.hdr -a $$opts{af_annots} -Ob -o $outfile.part && " .
+            "mv $outfile.part $outfile",%$opts);
     }
     closedir($dh) or error("close failed: $$opts{indir}");
 
@@ -234,7 +209,7 @@ sub run_roh
     for my $file (@files)
     {
         if ( -e "$file.txt.gz" ) { next; }
-        my @out = cmd("bcftools roh $$opts{roh_args} --AF-tag AF1KG $genmap $file -Orz -o $file.txt.gz.part 2>&1 | tee -a $file.log",%$opts);
+        my @out = cmd("bcftools roh --AF-tag AF1KG $genmap $file -Orz -o $file.txt.gz.part 2>&1 | tee -a $file.log",%$opts);
         for my $line (@out)
         {
             if ( !($line=~m{total/processed:\s+(\d+)/(\d+)}) ) { next; }
